from typing import Any, Dict, Iterator

from pygments.lexer import DelegatingLexer
from pygments.lexers import PythonLexer
from pygments.token import Name, _TokenType
from sphinx.application import Sphinx
from sphinx.highlighting import lexers
from tree_sitter import Language, Node, Parser
import tree_sitter_python


KNOWN_MODULES = {
    "ablate",
    "ablate.sources",
    "ablate.queries",
    "ablate.blocks",
    "ablate.exporters",
}


class TreeSitterOverlayLexer(PythonLexer):
    def __init__(self, **options: Any) -> None:
        super().__init__(**options)
        self.parser = Parser(Language(tree_sitter_python.language()))

    def get_tokens_unprocessed(
        self,
        text: str,
    ) -> Iterator[tuple[int, _TokenType, str]]:
        byte_text = text.encode("utf-8")
        tree = self.parser.parse(byte_text)
        spans: dict[tuple[int, str], _TokenType] = {}

        def classify(node: Node, value: str) -> _TokenType:
            p = node.parent
            gp = p.parent if p else None

            if (
                p
                and p.type == "attribute"
                and gp
                and gp.type == "call"
                and p == gp.child_by_field_name("function")
                and node == p.child_by_field_name("attribute")
            ):
                return Name.Function if value[0].islower() else Name.Class

            if p and p.type == "call" and node == p.child_by_field_name("function"):
                return Name.Function if value[0].islower() else Name.Class

            if p and p.type == "attribute":
                if node == p.child_by_field_name("object"):
                    return Name.Namespace
                if node == p.child_by_field_name("attribute"):
                    return Name.Namespace if value.islower() else Name.Class

            if p and "import" in p.type:
                return Name.Namespace

            return Name.Class if value[0].isupper() else Name

        def walk(node: Node) -> None:
            if node.type == "identifier":
                val = byte_text[node.start_byte : node.end_byte].decode(
                    "utf-8", errors="ignore"
                )
                spans.setdefault((node.start_byte, val), classify(node, val))
            for child in node.children:
                walk(child)

        walk(tree.root_node)

        for index, token, val in super().get_tokens_unprocessed(text):
            yield index, spans.get((index, val), token), val


class AblateLexer(DelegatingLexer):
    name = "ablate-python"
    aliases = ["ablate-python"]

    _KNOWN_MODULE_PARTS = {
        "ablate",
        "sources",
        "queries",
        "blocks",
        "exporters",
        "ablate.sources",
        "ablate.queries",
        "ablate.blocks",
        "ablate.exporters",
    }

    def __init__(self, **options: Any) -> None:
        super().__init__(PythonLexer, TreeSitterOverlayLexer, **options)

    def get_tokens_unprocessed(
        self,
        text: str,
    ) -> Iterator[tuple[int, _TokenType, str]]:
        for index, token, val in super().get_tokens_unprocessed(text):
            if token in {Name, Name.Namespace} and val in self._KNOWN_MODULE_PARTS:
                yield index, Name.Class, val
            else:
                yield index, token, val


def setup(app: Sphinx) -> Dict[str, Any]:
    lexers["python"] = AblateLexer(startinline=True)
    return {
        "version": "0.1.0",
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



from typing import List

from ablate.blocks import Text
from ablate.core.types import Run
from ablate.report import Report


def make_runs() -> List[Run]:
    return [
        Run(id="a", params={"model": "resnet", "seed": 1}, metrics={"accuracy": 0.7}),
        Run(id="b", params={"model": "resnet", "seed": 2}, metrics={"accuracy": 0.8}),
    ]


def test_report_add() -> None:
    report = Report(make_runs())
    report.add(Text("Test 1"), Text("Test 2"))
    assert len(report.blocks) == 2
    assert isinstance(report.blocks[0], Text)


def test_report_iadd() -> None:
    report = Report(make_runs())
    report += Text("Test")
    assert len(report.blocks) == 1
    assert isinstance(report.blocks[0], Text)


def test_report_add_operator() -> None:
    report = Report(make_runs())
    new_report = report + Text("Test")
    assert len(report.blocks) == 0
    assert len(new_report.blocks) == 1
    assert isinstance(new_report.blocks[0], Text)



from pathlib import Path
from typing import List

from ablate.blocks import (
    AbstractFigureBlock,
    AbstractTableBlock,
    AbstractTextBlock,
    MetricPlot,
    Text,
)
from ablate.blocks.text_blocks import _Heading
from ablate.core.types import Run
from ablate.exporters.abstract_exporter import AbstractExporter
from ablate.report import Report

from .utils import HEADING_LEVELS, hash_dataframe, render_metric_plot


class Markdown(AbstractExporter):
    def __init__(
        self,
        output_path: str = "report.md",
        assets_dir: str | None = None,
        export_csv: bool = False,
    ) -> None:
        """Export the report as a markdown file.

        Args:
            output_path: The path to the output markdown file. Defaults to "report.md".
            assets_dir: The directory to store the assets (figures, etc.). If None,
                defaults to the parent directory of the output file with a ".ablate"
                subdirectory. Defaults to None.
            export_csv: Whether to export tables and plots as CSV files.
                Defaults to False.
        """
        self.output_path = Path(output_path)
        self.assets_dir = (
            Path(assets_dir) if assets_dir else self.output_path.parent / ".ablate"
        )
        self.assets_dir.mkdir(exist_ok=True)
        self.export_csv = export_csv

    def export(self, report: Report) -> None:
        content = self.render_blocks(report)
        with self.output_path.open("w", encoding="utf-8") as f:
            for block_output in content:
                f.write(block_output)
                f.write("\n\n")

    def render_text(self, block: AbstractTextBlock, runs: List[Run]) -> str:
        if isinstance(block, Text):
            return block.build(runs)
        if isinstance(block, _Heading):
            return f"{'#' * HEADING_LEVELS[type(block)]} {block.build(runs)}"
        raise NotImplementedError(f"Unsupported text block: '{type(block)}'.")

    def render_table(self, block: AbstractTableBlock, runs: List[Run]) -> str:
        df = block.build(runs)
        if self.export_csv:
            df.to_csv(
                self.assets_dir / f"{type(block).__name__}_{hash_dataframe(df)}.csv",
                index=False,
            )
        return df.to_markdown(index=False)

    def render_figure(self, block: AbstractFigureBlock, runs: List[Run]) -> str:
        if not isinstance(block, MetricPlot):
            raise NotImplementedError(f"Unsupported figure block: '{type(block)}'.")

        if self.export_csv:
            df = block.build(runs)
            df.to_csv(
                self.assets_dir / f"{type(block).__name__}_{hash_dataframe(df)}.csv",
                index=False,
            )

        filename = render_metric_plot(block, runs, self.assets_dir)
        if filename is None:
            return (
                f"*No data available for {', '.join(m.label for m in block.metrics)}*"
            )
        return f"![{filename}](.ablate/{filename})"



from unittest.mock import MagicMock, patch

import pytest

from ablate.sources import ClearML


@patch("clearml.Task.get_task")
@patch("clearml.Task.query_tasks")
def test_clearml_loads_metrics_and_temporal_data(
    mock_query_tasks: MagicMock,
    mock_get_task: MagicMock,
) -> None:
    mock_query_tasks.return_value = ["clearml-001"]

    mock_task = MagicMock()
    mock_task.id = "clearml-001"
    mock_task.get_parameters.return_value = {"lr": 0.001, "epochs": 10}
    mock_task.get_reported_scalars.return_value = {
        "title": {
            "accuracy": {"x": [1, 2], "y": [0.85, 0.9]},
            "loss": {"x": [1, 2], "y": [0.25, 0.2]},
        }
    }

    mock_get_task.return_value = mock_task

    source = ClearML(project_name="example")
    runs = source.load()

    assert len(runs) == 1
    r = runs[0]
    assert r.id == "clearml-001"
    assert r.params == {"lr": 0.001, "epochs": 10}
    assert r.metrics == {"accuracy": 0.9, "loss": 0.2}
    assert r.temporal == {
        "accuracy": [(1, 0.85), (2, 0.9)],
        "loss": [(1, 0.25), (2, 0.2)],
    }


@patch.dict("sys.modules", {"clearml": None})
def test_import_error_if_clearml_not_installed() -> None:
    with pytest.raises(ImportError, match="ClearML source requires `clearml`"):
        ClearML(project_name="fail")



from pathlib import Path
import re
from typing import List

import matplotlib.pyplot as plt
import pytest

from ablate.blocks import H1, H2, MetricPlot, Table, Text
from ablate.core.types import Run
from ablate.exporters import Markdown
from ablate.queries import Metric, Param
from ablate.report import Report

from .utils import DummyFigureBlock, DummyTextBlock


class DummyBlock:
    pass


@pytest.fixture
def runs() -> list[Run]:
    return [
        Run(
            id="run1",
            params={"model": "resnet"},
            metrics={"accuracy": 0.8},
            temporal={"accuracy": [(0, 0.5), (1, 0.8)]},
        ),
        Run(
            id="run2",
            params={"model": "resnet"},
            metrics={"accuracy": 0.9},
            temporal={"accuracy": [(0, 0.6), (1, 0.9)]},
        ),
    ]


def test_export_text_blocks(tmp_path: Path, runs: List[Run]) -> None:
    report = Report(runs).add(H1("Heading 1"), Text("Some paragraph text."))
    out_path = tmp_path / "report.md"
    Markdown(output_path=str(out_path)).export(report)

    content = out_path.read_text()
    assert "# Heading 1" in content
    assert "Some paragraph text." in content


def test_export_table_block(tmp_path: Path, runs: List[Run]) -> None:
    table = Table(
        columns=[
            Param("model", label="Model"),
            Metric("accuracy", direction="max", label="Accuracy"),
        ]
    )
    report = Report(runs).add(table)
    out_path = tmp_path / "report.md"
    Markdown(output_path=str(out_path)).export(report)

    content = out_path.read_text().replace("\r\n", "\n")
    assert re.search(r"\|\s*Model\s*\|\s*Accuracy\s*\|", content)
    pattern = (
        r"\|\s*resnet\s*\|\s*0\.8\s*\|\s*\n"
        r"\|\s*resnet\s*\|\s*0\.9\s*\|"
    )
    assert re.search(pattern, content)


def test_export_figure_block(tmp_path: Path, runs: List[Run]) -> None:
    plot = MetricPlot(Metric("accuracy", direction="max"), identifier=Param("model"))
    report = Report(runs).add(plot)
    out_path = tmp_path / "report.md"
    exporter = Markdown(output_path=str(out_path))
    exporter.export(report)

    content = out_path.read_text()
    assert "![MetricPlot_" in content
    asset_path = tmp_path / ".ablate"
    assert asset_path.exists()
    files = list(asset_path.glob("MetricPlot_*.png"))
    assert len(files) == 1
    img = plt.imread(files[0])
    assert img.shape[-1] in {3, 4}


def test_export_figure_block_empty(tmp_path: Path) -> None:
    empty_run = Run(id="x", params={}, metrics={}, temporal={})
    plot = MetricPlot(Metric("accuracy", direction="max"), identifier=Param("model"))
    report = Report([empty_run]).add(plot)
    out_path = tmp_path / "report.md"
    Markdown(output_path=str(out_path)).export(report)

    content = out_path.read_text()
    assert "*No data available for accuracy*" in content


def test_unknown_block_raises(tmp_path: Path, runs: List[Run]) -> None:
    report = Report(runs)
    report += DummyBlock()  # type: ignore[arg-type]
    with pytest.raises(ValueError, match="Unknown block type"):
        Markdown(output_path=str(tmp_path / "out.md")).export(report)


def test_unsupported_figure_block_raises(tmp_path: Path, runs: List[Run]) -> None:
    report = Report(runs).add(DummyFigureBlock())
    exporter = Markdown(output_path=str(tmp_path / "out.md"))
    with pytest.raises(NotImplementedError, match="Unsupported figure block"):
        exporter.export(report)


def test_unsupported_text_block_raises(tmp_path: Path, runs: List[Run]) -> None:
    report = Report(runs).add(DummyTextBlock("oops"))
    exporter = Markdown(output_path=str(tmp_path / "out.md"))
    with pytest.raises(NotImplementedError, match="Unsupported text block"):
        exporter.export(report)


def test_block_level_runs_override_global(tmp_path: Path, runs: List[Run]) -> None:
    scoped_runs = [runs[0]]
    report = Report(runs).add(
        Text("global"),
        Table([Param("model"), Metric("accuracy", "max")], runs=scoped_runs),
    )
    out_path = tmp_path / "report.md"
    Markdown(output_path=str(out_path)).export(report)
    content = out_path.read_text()
    assert "resnet" in content
    assert content.count("resnet") == 1


def test_export_heading_variants(tmp_path: Path, runs: List[Run]) -> None:
    report = Report(runs).add(H2("Section Title"))
    out_path = tmp_path / "headings.md"
    Markdown(output_path=str(out_path)).export(report)
    content = out_path.read_text()
    assert "## Section Title" in content


def test_export_table_block_with_csv(tmp_path: Path, runs: List[Run]) -> None:
    table = Table(
        columns=[Param("model"), Metric("accuracy", direction="max")],
    )
    report = Report(runs).add(table)
    out_path = tmp_path / "report.md"
    exporter = Markdown(output_path=str(out_path), export_csv=True)
    exporter.export(report)

    content = out_path.read_text()
    assert "resnet" in content
    asset_dir = tmp_path / ".ablate"
    csv_files = list(asset_dir.glob("Table_*.csv"))
    assert len(csv_files) == 1
    csv_content = csv_files[0].read_text()
    assert "model,accuracy" in csv_content
    assert "resnet,0.8" in csv_content


def test_export_figure_block_with_csv(tmp_path: Path, runs: List[Run]) -> None:
    plot = MetricPlot(Metric("accuracy", direction="max"), identifier=Param("model"))
    report = Report(runs).add(plot)
    out_path = tmp_path / "report.md"
    exporter = Markdown(output_path=str(out_path), export_csv=True)
    exporter.export(report)

    asset_dir = tmp_path / ".ablate"
    png_files = list(asset_dir.glob("MetricPlot_*.png"))
    assert len(png_files) == 1
    csv_files = list(asset_dir.glob("MetricPlot_*.csv"))
    assert len(csv_files) == 1
    csv_content = csv_files[0].read_text()
    assert "step,value,metric,run,run_id" in csv_content
    assert "0,0.5,accuracy,resnet,run1" in csv_content
    assert "1,0.9,accuracy,resnet,run2" in csv_content



from typing import List

import pandas as pd

from ablate.core.types import Run

from .abstract_source import AbstractSource


class WandB(AbstractSource):
    def __init__(self, project: str, entity: str | None = None) -> None:
        """Weights & Biases (WandB) source for loading runs from a WandB project.

        Args:
            project: The name of the WandB project to load runs from.
            entity: Optional WandB entity (team or user). If None, uses the default
                entity from the WandB configuration. Defaults to None.

        Raises:
            ImportError: If the `wandb` package is not installed.
        """

        try:
            import wandb
        except ImportError as e:
            raise ImportError(
                "Wandb source requires `wandb`. "
                "Install via `pip install ablate[wandb]`."
            ) from e
        self.project = project
        self.entity = entity or wandb.Api().default_entity
        self.api = wandb.Api()

    def load(self) -> List[Run]:
        runs = self.api.runs(f"{self.entity}/{self.project}")
        records = []
        for r in runs:
            params = dict(r.config)
            metrics = {
                k: v for k, v in r.summary.items() if isinstance(v, (int, float))
            }
            temporal = {}
            for key in metrics:
                df = r.history(keys=[key])
                if (
                    isinstance(df, pd.DataFrame)
                    and key in df.columns
                    and "_step" in df.columns
                ):
                    temporal[key] = list(zip(df["_step"], df[key], strict=False))

            records.append(
                Run(id=r.id, params=params, metrics=metrics, temporal=temporal)
            )
        return records



import itertools
from typing import Any, Dict, List

import numpy as np

from ablate.core.types import Run

from .abstract_source import AbstractSource


class Mock(AbstractSource):
    def __init__(
        self,
        grid: Dict[str, List[Any]],
        num_seeds: int = 1,
        steps: int = 25,
    ) -> None:
        """Mock source for generating runs based on a grid of hyperparameters.

        For each run, `accuracy`, `f1`, and `loss` metrics are randomly generated.

        Args:
            grid: Dictionary mapping parameter names to lists of values.
            num_seeds: Number of runs to generate per config with different seeds.
                Defaults to 1.
            steps: Number of steps to use for temporal metrics.
                Defaults to 25.
        """
        self.grid = grid
        self.num_seeds = num_seeds
        self.steps = steps

    def _generate_runs(self, param_dict: Dict[str, Any], idx: int) -> List[Run]:
        runs: List[Run] = []
        for local_seed in range(self.num_seeds):
            global_seed = idx * self.num_seeds + local_seed
            rng = np.random.default_rng(global_seed)
            steps = np.arange(1, self.steps + 1, dtype=np.int32)

            progress = steps / self.steps
            growth = 1 / (1 + np.exp(-6 * (progress - 0.5)))
            _accc = 0.6 + 0.35 * growth + rng.normal(0, 0.01, len(steps))
            accc = np.clip(_accc, 0.0, 1.0)
            _f1c = accc - rng.uniform(0.01, 0.05, len(steps))
            f1c = np.clip(_f1c, 0.0, 1.0)
            _lossc = 1.5 / (steps + 1) + rng.uniform(0.01, 0.05, len(steps))
            lossc = np.clip(_lossc, 1e-4, None)

            accuracy = float(np.round(accc[-1], 4))
            f1 = float(np.round(f1c[-1], 4))
            loss = float(np.round(lossc[-1], 4))

            params = {
                **{k: str(v) for k, v in param_dict.items()},
                "seed": str(local_seed),
            }
            rid = f"{'_'.join(f'{k}={v}' for k, v in params.items())}"
            m = {"accuracy": accuracy, "f1": f1, "loss": loss}
            t = {
                "accuracy": list(zip(steps.tolist(), accc.tolist(), strict=False)),
                "f1": list(zip(steps.tolist(), f1c.tolist(), strict=False)),
                "loss": list(zip(steps.tolist(), lossc.tolist(), strict=False)),
            }
            runs.append(Run(id=rid, params=params, metrics=m, temporal=t))
        return runs

    def load(self) -> List[Run]:
        param_names = list(self.grid.keys())
        param_product = list(itertools.product(*[self.grid[k] for k in param_names]))
        runs: List[Run] = []

        for idx, product in enumerate(param_product):
            param_dict = dict(zip(param_names, product, strict=False))
            runs.extend(self._generate_runs(param_dict, idx))

        return runs



from __future__ import annotations

from collections import defaultdict
from copy import deepcopy
import hashlib
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Set, Tuple, Union

from ablate.core.types import GroupedRun, Run

from .grouped_query import GroupedQuery


if TYPE_CHECKING:  # pragma: no cover
    from .selectors import AbstractMetric, AbstractParam


class Query:
    def __init__(self, runs: List[Run]) -> None:
        """Query interface for manipulating runs in a functional way.

        All methods operate on a shallow copy of the runs in the query, so the original
        runs are not modified and assumed to be immutable.

        Args:
            runs: List of runs to be queried.
        """
        self._runs = runs

    def filter(self, fn: Callable[[Run], bool]) -> Query:
        """Filter the runs in the query based on a predicate function.

        Args:
            fn: Predicate function that takes in a run and returns a boolean value.

        Returns:
            A new query with the runs that satisfy the predicate function.
        """
        return Query([r for r in self._runs[:] if fn(r)])

    def map(self, fn: Callable[[Run], Run]) -> Query:
        """Apply a function to each run in the query.

        This function is intended to be used for modifying the runs in the query. The
        function should return a new run object as the original run is not modified.

        Args:
            fn: Function that takes in a run and returns a new run object.

        Returns:
            A new query with the modified runs.
        """
        return Query([fn(r) for r in deepcopy(self._runs)])

    def sort(self, key: AbstractMetric, ascending: bool = False) -> Query:
        """Sort the runs in the query based on a metric.

        Args:
            key: Metric to sort the runs by.
            ascending: Whether to sort in ascending order.
                Defaults to False (descending order).

        Returns:
            A new query with the runs sorted by the specified metric.
        """
        return Query(sorted(self._runs[:], key=key, reverse=not ascending))

    def project(self, selectors: Union[AbstractParam, List[AbstractParam]]) -> Query:
        """Project the parameter space of the runs in the query to a subset of
        parameters only including the specified selectors.

        This function is intended to be used for reducing the dimensionality of the
        parameter space and therefore operates on a deep copy of the runs in the query.

        Args:
            selectors: Selector or list of selectors to project the runs by.

        Returns:
            A new query with the projected runs.
        """
        if not isinstance(selectors, list):
            selectors = [selectors]

        names = {s.name for s in selectors}
        projected: List[Run] = []

        for run in deepcopy(self._runs):
            run.params = {k: v for k, v in run.params.items() if k in names}
            projected.append(run)

        return Query(projected)

    def groupby(
        self,
        selectors: Union[AbstractParam, List[AbstractParam]],
    ) -> GroupedQuery:
        """Group the runs in the query by one or more selectors.

        Args:
            selectors: Selector or list of selectors to group the runs by.

        Returns:
            A grouped query containing the grouped runs.
        """
        if not isinstance(selectors, list):
            selectors = [selectors]

        def key_fn(run: Run) -> Tuple[Any, ...]:
            return tuple(selector(run) for selector in selectors)

        groups = defaultdict(list)
        for run in self._runs:
            groups[key_fn(run)].append(run)

        grouped = [
            GroupedRun(
                key="+".join(selector.name for selector in selectors),
                value="|".join(map(str, k)),
                runs=v,
            )
            for k, v in groups.items()
        ]
        return GroupedQuery(grouped)

    def groupdiff(
        self,
        selectors: Union[AbstractParam, List[AbstractParam]],
    ) -> GroupedQuery:
        """Group the runs in the query by one or more selectors, excluding the keys.
        This is similar to `groupby` but it excludes the specified keys from the
        grouping key.

        Args:
            selectors: Selector or list of selectors to exclude from the grouping key.

        Returns:
            A grouped query containing the grouped runs.
        """
        if not isinstance(selectors, list):
            selectors = [selectors]

        def exclude_keys(
            d: Dict[str, Any],
            keys: Set[str],
        ) -> Tuple[Tuple[str, Any], ...]:
            return tuple(sorted((k, v) for k, v in d.items() if k not in keys))

        exclude_names = {s.name for s in selectors}
        groups = defaultdict(list)
        for run in self._runs:
            key = exclude_keys(run.params, exclude_names)
            groups[key].append(run)

        def _hash_key(kv: Tuple[Tuple[str, Any], ...]) -> str:
            raw = ",".join(f"{k}={v}" for k, v in sorted(kv))
            return hashlib.md5(raw.encode()).hexdigest()[:8]

        grouped = [
            GroupedRun(
                key="-".join(s.name for s in selectors),
                value=_hash_key(k),
                runs=runs,
            )
            for k, runs in groups.items()
        ]
        return GroupedQuery(grouped)

    def head(self, n: int) -> Query:
        """Get the first n runs in the query.

        Args:
            n: Number of runs to return.

        Returns:
            A new query with the first n runs.
        """
        return Query(self._runs[:n])

    def tail(self, n: int) -> Query:
        """Get the last n runs in the query.

        Args:
            n: Number of runs to return.

        Returns:
            A new query with the last n runs.
        """
        return Query(self._runs[-n:])

    def topk(self, metric: AbstractMetric, k: int) -> Query:
        """Get the top k runs in the query based on a metric.

        Args:
            metric: Metric to sort the runs by.
            k: Number of top runs to return.

        Returns:
            A new query with the top k runs based on the specified metric.
        """
        return self.sort(metric, ascending=metric.direction == "min").head(k)

    def bottomk(self, metric: AbstractMetric, k: int) -> Query:
        """Get the bottom k runs in the query based on a metric.

        Args:
            metric: Metric to sort the runs by.
            k: Number of bottom runs to return.

        Returns:
            A new query with the bottom k runs based on the specified metric.
        """
        return self.sort(metric, ascending=metric.direction == "max").head(k)

    def all(self) -> List[Run]:
        """Collect all runs in the query.

        Returns:
            A list of all runs in the query.
        """
        return deepcopy(self._runs)

    def copy(self) -> Query:
        """Obtain a shallow copy of the query.

        Returns:
            A new query with the same runs as the original query.
        """
        return Query(self._runs[:])

    def deepcopy(self) -> Query:
        """Obtain a deep copy of the query.

        Returns:
            A new query with deep copies of the runs in the original query.
        """
        return Query(deepcopy(self._runs))

    def __len__(self) -> int:
        """Get the number of runs in the query.

        Returns:
            The number of runs in the query.
        """
        return len(self._runs)



from typing import List
from unittest.mock import patch

import matplotlib.pyplot as plt
import pandas as pd
import pytest

from ablate import Report
from ablate.blocks import H2, MetricPlot, Table, Text
from ablate.core.types import Run
from ablate.exporters.notebook_exporter import Notebook, running_in_notebook
from ablate.queries import Metric, Param

from .utils import DummyFigureBlock, DummyTextBlock


@pytest.fixture
def runs() -> List[Run]:
    return [
        Run(
            id="r1",
            params={"model": "resnet"},
            metrics={"acc": 0.8},
            temporal={"acc": [(0, 0.5), (1, 0.8)]},
        ),
        Run(
            id="r2",
            params={"model": "resnet"},
            metrics={"acc": 0.9},
            temporal={"acc": [(0, 0.6), (1, 0.9)]},
        ),
    ]


def test_render_text_block(runs: List[Run]) -> None:
    block = Text("Hello *world*")
    with patch("IPython.display.display") as mock_display:
        Notebook().render_text(block, runs)
        assert mock_display.call_count == 1
        assert "Hello" in mock_display.call_args[0][0].data


def test_render_heading_block(runs: List[Run]) -> None:
    block = H2("Section")
    with patch("IPython.display.display") as mock_display:
        Notebook().render_text(block, runs)
        assert mock_display.call_count == 1
        assert mock_display.call_args[0][0].data.startswith("##")


def test_render_table_block(runs: List[Run]) -> None:
    table = Table(columns=[Param("model"), Metric("acc", direction="max")])
    with patch("IPython.display.display") as mock_display:
        Notebook().render_table(table, runs)
        df = mock_display.call_args[0][0]
        assert isinstance(df, pd.DataFrame)
        assert "model" in df.columns
        assert len(df) == 2


def test_render_metric_plot(runs: List[Run]) -> None:
    plot = MetricPlot(Metric("acc", direction="max"), identifier=Param("model"))
    with patch("IPython.display.display") as mock_display:
        Notebook().render_figure(plot, runs)
        fig = mock_display.call_args[0][0]
        assert isinstance(fig, plt.Figure)


def test_render_empty_plot() -> None:
    empty_run = Run(id="x", params={}, metrics={}, temporal={})
    plot = MetricPlot(Metric("acc", direction="max"), identifier=Param("model"))
    with patch("IPython.display.display") as mock_display:
        Notebook().render_figure(plot, [empty_run])
        text = mock_display.call_args[0][0]
        assert "*No data available for acc*" in text.data


def test_running_in_notebook_true(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr("sys.argv", ["ipykernel_launcher"])
    assert running_in_notebook() is True


def test_running_in_notebook_false(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr("sys.argv", ["some_script.py"])
    assert running_in_notebook() is False


def test_notebook_import_error() -> None:
    with (
        patch.dict("sys.modules", {"IPython.display": None}),
        pytest.raises(ImportError, match="requires `jupyter`"),
    ):
        Notebook()


def test_export_runs(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr("sys.argv", ["ipykernel_launcher"])
    report = Report([])
    with patch.object(Notebook, "render_blocks", return_value=[]) as mock_render:
        Notebook().export(report)
        mock_render.assert_called_once_with(report)


def test_export_outside_notebook(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr("sys.argv", ["script.py"])

    with pytest.raises(RuntimeError, match="only be used inside a Jupyter notebook"):
        Notebook().export(Report([]))


def test_render_text_not_implemented(runs: List[Run]) -> None:
    dummy = DummyTextBlock("oops")
    with pytest.raises(NotImplementedError):
        Notebook().render_text(dummy, runs)


def test_render_figure_not_implemented(runs: List[Run]) -> None:
    dummy = DummyFigureBlock()
    with pytest.raises(NotImplementedError):
        Notebook().render_figure(dummy, runs)



from .abstract_source import AbstractSource
from .autrainer_source import Autrainer
from .clearml_source import ClearML
from .mlflow_source import MLflow
from .mock_source import Mock
from .tensorboard_source import TensorBoard
from .wandb_source import WandB


__all__ = [
    "AbstractSource",
    "Autrainer",
    "ClearML",
    "MLflow",
    "Mock",
    "TensorBoard",
    "WandB",
]



from abc import ABC, abstractmethod
from typing import List

from ablate.core.types import Run


class AbstractSource(ABC):
    @abstractmethod
    def load(self) -> List[Run]:
        """Load the data from the source.

        Returns:
            A list of runs with their parameters, metrics, and optionally temporal data.
        """



from pydantic import ValidationError
import pytest

from ablate.core.types import GroupedRun, Run


def test_run() -> None:
    run = Run(id="test_run", params={"param1": 1}, metrics={"metric1": 0.5})
    assert run.id == "test_run"
    assert run.params == {"param1": 1}
    assert run.metrics == {"metric1": 0.5}
    assert run.temporal == {}


def test_temporal_data() -> None:
    run = Run(
        id="test_run",
        params={"param1": 1},
        metrics={"metric1": 0.5},
        temporal={"metric1": [(0, 0.0), (1, 1.0), (2, 2.0)]},
    )
    assert run.temporal == {"metric1": [(0, 0.0), (1, 1.0), (2, 2.0)]}
    assert run.temporal["metric1"][0] == (0, 0.0)


def test_invalid_metrics_type() -> None:
    with pytest.raises(ValidationError):
        Run(id="bad", params={}, metrics="not a dict")  # type: ignore[arg-type]


def test_grouped_run() -> None:
    run1 = Run(id="run1", params={"param1": 1}, metrics={"metric1": 0.5})
    run2 = Run(id="run2", params={"param2": 2}, metrics={"metric2": 0.8})
    grouped_run = GroupedRun(key="group_key", value="group_value", runs=[run1, run2])
    assert grouped_run.key == "group_key"
    assert grouped_run.value == "group_value"
    assert len(grouped_run.runs) == 2
    assert grouped_run.runs[0].id == "run1"
    assert grouped_run.runs[1].id == "run2"


def test_invalid_runs_type() -> None:
    with pytest.raises(ValidationError):
        GroupedRun(key="group_key", value="group_value", runs=["not a run"])  # type: ignore[list-item]


def test_run_roundtrip_serialization() -> None:
    run = Run(
        id="run1",
        params={"x": 1},
        metrics={"acc": 0.9},
        temporal={"loss": [(0, 0.1), (1, 0.05)]},
    )
    data = run.model_dump()
    recovered = Run(**data)
    assert recovered == run



from abc import ABC
from typing import List

from ablate.core.types import Run

from .abstract_block import AbstractBlock


class AbstractTextBlock(AbstractBlock, ABC):
    def __init__(self, text: str, runs: List[Run] | None = None) -> None:
        """Block containing styled text for a report.

        Args:
            text: The text content of the block.
            runs: Optional list of runs to be used for the block instead of the default
                runs from the report. Defaults to None.
        """
        super().__init__(runs)
        self.text = text

    def build(self, runs: List[Run]) -> str:
        return self.text.strip()


class _Heading(AbstractTextBlock): ...


class H1(_Heading): ...


class H2(_Heading): ...


class H3(_Heading): ...


class H4(_Heading): ...


class H5(_Heading): ...


class H6(_Heading): ...


class Text(AbstractTextBlock): ...



from typing import List

from ablate.core.types import Run

from .abstract_source import AbstractSource


class ClearML(AbstractSource):
    def __init__(self, project_name: str) -> None:
        """ClearML source for loading runs from a ClearML server.

        Args:
            project_name: The name of the ClearML project to load runs from.

        Raises:
            ImportError: If the `clearml` package is not installed.
        """
        try:
            from clearml import Task  # noqa: F401
        except ImportError as e:
            raise ImportError(
                "ClearML source requires `clearml`. "
                "Install via `pip install ablate[clearml]`."
            ) from e
        self.project_name = project_name

    def load(self) -> List[Run]:
        from clearml import Task

        task_ids = Task.query_tasks(project_name=self.project_name)
        records = []

        for task_id in task_ids:
            t = Task.get_task(task_id=task_id)
            params = t.get_parameters() or {}
            scalars = t.get_reported_scalars() or {}

            metrics = {}
            temporal = {}

            for _, series in scalars.items():
                for name, values in series.items():
                    if not values or "x" not in values or "y" not in values:
                        continue  # pragma: no cover
                    x, y = values["x"], values["y"]
                    if isinstance(x, list) and isinstance(y, list) and len(x) == len(y):
                        temporal[name] = list(zip(x, y, strict=False))
                        metrics[name] = float(y[-1])

            records.append(
                Run(id=t.id, params=params, metrics=metrics, temporal=temporal)
            )

        return records



from .grouped_query import GroupedQuery
from .query import Query
from .selectors import (
    AbstractMetric,
    AbstractParam,
    AbstractSelector,
    Id,
    Metric,
    Param,
    TemporalMetric,
)


__all__ = [
    "AbstractMetric",
    "AbstractParam",
    "AbstractSelector",
    "GroupedQuery",
    "Id",
    "Metric",
    "Param",
    "Query",
    "TemporalMetric",
]



from pathlib import Path
from typing import TYPE_CHECKING

import pandas as pd
import pytest
import yaml

from ablate.sources import Autrainer
from ablate.sources.autrainer_source import flatten_autrainer_config


if TYPE_CHECKING:
    from ablate.core.types import Run


def write_yaml(path: Path, content: dict) -> None:
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(yaml.dump(content))


def write_csv(path: Path, content: dict) -> None:
    df = pd.DataFrame(content)
    df.to_csv(path, index=False)


def make_dummy_run(path: Path) -> None:
    write_yaml(
        path / ".hydra" / "config.yaml",
        {
            "model": {
                "id": "MyModel",
                "hidden": 128,
                "nested": {"id": "Submodule", "layers": 2},
            },
            "optimizer": {"lr": 0.01},
        },
    )
    write_yaml(
        path / "_best" / "dev.yaml",
        {"accuracy": {"all": 0.85}, "loss": {"all": 0.3}, "iteration": 5},
    )
    write_yaml(
        path / "_test" / "test_holistic.yaml",
        {"accuracy": {"all": 0.8}, "loss": {"all": 0.35}, "iteration": 5},
    )
    write_csv(
        path / "metrics.csv",
        {
            "iteration": [1, 2, 3],
            "accuracy": [0.7, 0.75, 0.8],
            "loss": [0.4, 0.38, 0.36],
        },
    )


def test_raises_if_experiment_missing(tmp_path: Path) -> None:
    with pytest.raises(FileNotFoundError, match="No autrainer experiment found"):
        Autrainer(results_dir=str(tmp_path), experiment_id="no_such_id")


def test_loads_single_run_correctly(tmp_path: Path) -> None:
    exp_id = "exp-001"
    run_path = tmp_path / exp_id / "training" / "run0"
    make_dummy_run(run_path)

    source = Autrainer(results_dir=str(tmp_path), experiment_id=exp_id)
    runs = source.load()

    assert len(runs) == 1
    r: Run = runs[0]
    assert r.id == "run0"

    # Flattened params
    assert r.params["model"] == "MyModel"
    assert r.params["model.nested"] == "Submodule"
    assert r.params["model.hidden"] == 128
    assert r.params["model.nested.layers"] == 2
    assert r.params["optimizer.lr"] == 0.01

    # Merged dev/test metrics
    assert r.metrics == {
        "accuracy": 0.85,
        "loss": 0.3,
        "test_accuracy": 0.8,
        "test_loss": 0.35,
    }

    # Temporal metrics
    assert r.temporal == {
        "accuracy": [(1, 0.7), (2, 0.75), (3, 0.8)],
        "loss": [(1, 0.4), (2, 0.38), (3, 0.36)],
    }


def test_skips_non_directories(tmp_path: Path) -> None:
    exp_id = "exp-002"
    base = tmp_path / exp_id / "training"
    run_dir = base / "run_dir"
    run_file = base / "README.txt"

    make_dummy_run(run_dir)
    run_file.write_text("This is not a run directory")

    source = Autrainer(results_dir=str(tmp_path), experiment_id=exp_id)
    runs = source.load()

    assert len(runs) == 1
    assert runs[0].id == "run_dir"


def test_flatten_config_yields_id_only_dict() -> None:
    config = {"id": "only"}
    result = dict(flatten_autrainer_config(config))
    assert result == {"": "only"}


def test_flatten_config_with_list_of_dicts() -> None:
    config = [
        {"id": "first"},
        {"name": "second", "id": "second_id"},
    ]
    result = dict(flatten_autrainer_config(config))

    assert result == {"0": "first", "1": "second_id", "1.name": "second"}



import hashlib
from pathlib import Path
from typing import List

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns

from ablate.blocks import H1, H2, H3, H4, H5, H6, MetricPlot
from ablate.core.types.runs import Run


HEADING_LEVELS = {H1: 1, H2: 2, H3: 3, H4: 4, H5: 5, H6: 6}


def apply_default_plot_style() -> None:
    sns.set_style("whitegrid")
    sns.set_context("paper", font_scale=0.8)
    sns.set_palette("muted")
    plt.rcParams["figure.dpi"] = 300


def hash_dataframe(df: pd.DataFrame) -> str:
    return hashlib.md5(df.to_csv(index=False).encode("utf-8")).hexdigest()[:12]


def create_metric_plot(df: pd.DataFrame, label: str) -> plt.Figure:
    apply_default_plot_style()
    fig, ax = plt.subplots()
    sns.lineplot(
        data=df,
        x="step",
        y="value",
        hue="run",
        style="metric" if df["metric"].nunique() > 1 else None,
        ax=ax,
    )
    ax.set_xlabel("Step")
    ax.set_ylabel("Value")
    ax.legend(title=label, loc="best", frameon=False)
    plt.tight_layout()
    return fig


def render_metric_plot(
    block: MetricPlot,
    runs: List[Run],
    output_dir: Path,
) -> str | None:
    df = block.build(runs)
    if df.empty:
        return None

    fig = create_metric_plot(df, block.identifier.label)
    filename = f"{type(block).__name__}_{hash_dataframe(df)}.png"
    fig.savefig(output_dir / filename)
    plt.close(fig)
    return filename



from typing import List

import pytest

from ablate.core.types import Run
from ablate.sources import Mock


@pytest.fixture
def mock_runs() -> List[Run]:
    source = Mock(
        grid={"model": ["resnet", "vgg"], "lr": [0.01, 0.001]},
        num_seeds=2,
        steps=10,
    )
    return source.load()


def test_mock_source_run_count(mock_runs: List[Run]) -> None:
    assert len(mock_runs) == 8


def test_mock_source_metrics_in_bounds(mock_runs: List[Run]) -> None:
    for run in mock_runs:
        metrics = run.metrics
        assert "accuracy" in metrics
        assert "f1" in metrics
        assert "loss" in metrics

        assert 0.0 <= metrics["accuracy"] <= 1.0
        assert 0.0 <= metrics["f1"] <= 1.0
        assert metrics["loss"] > 0.0


def test_mock_source_temporal_length(mock_runs: List[Run]) -> None:
    for run in mock_runs:
        for key in ["accuracy", "f1", "loss"]:
            assert key in run.temporal
            assert len(run.temporal[key]) == 10



from unittest.mock import MagicMock, patch

import pandas as pd
import pytest

from ablate.sources import WandB


@patch("wandb.Api")
def test_wandb_loads_metrics_and_temporal_data(mock_api_class: MagicMock) -> None:
    mock_api = mock_api_class.return_value
    mock_run = MagicMock()
    mock_run.id = "run-123"
    mock_run.config = {"lr": 0.01, "batch_size": 32}
    mock_run.summary = {"accuracy": 0.92, "loss": 0.1}

    mock_history_df = pd.DataFrame(
        {
            "_step": [1, 2],
            "accuracy": [0.89, 0.92],
        }
    )
    mock_run.history.return_value = mock_history_df
    mock_api.runs.return_value = [mock_run]

    source = WandB(project="my-project", entity="my-entity")
    runs = source.load()

    assert len(runs) == 1
    r = runs[0]
    assert r.id == "run-123"
    assert r.params == {"lr": 0.01, "batch_size": 32}
    assert r.metrics == {"accuracy": 0.92, "loss": 0.1}
    assert r.temporal == {"accuracy": [(1, 0.89), (2, 0.92)]}


@patch("wandb.Api")
def test_wandb_ignores_missing_history_keys(mock_api_class: MagicMock) -> None:
    mock_api = mock_api_class.return_value
    mock_run = MagicMock()
    mock_run.id = "run-xyz"
    mock_run.config = {}
    mock_run.summary = {"metric_logged": 0.5}

    mock_run.history.return_value = pd.DataFrame(
        {
            "_step": [1, 2],
            "some_other_metric": [0.3, 0.4],
        }
    )

    mock_api.runs.return_value = [mock_run]

    source = WandB(project="dummy", entity="dummy")
    runs = source.load()

    r = runs[0]
    assert r.metrics == {"metric_logged": 0.5}
    assert r.temporal == {}


@patch.dict("sys.modules", {"wandb": None})
def test_import_error_if_wandb_not_installed() -> None:
    with pytest.raises(ImportError, match="Wandb source requires `wandb`"):
        WandB(project="dummy")



from abc import ABC, abstractmethod
from operator import eq, ge, gt, le, lt, ne
from typing import Any, Callable, Literal

from ablate.core.types import Run


class AbstractSelector(ABC):
    def __init__(self, name: str, label: str | None = None) -> None:
        """Abstract class for selecting runs based on a specific attribute.

        Args:
            name: Name of the attribute to select on.
            label: Optional label for displaying purposes. If None, defaults to `name`.
                Defaults to None.
        """
        self.name = name
        self.label = label or name

    @abstractmethod
    def __call__(self, run: Run) -> Any: ...

    def _cmp(self, op: Callable[[Any, Any], bool], other: Any) -> Callable[[Run], bool]:
        return lambda run: op(self.__call__(run), other)

    def __eq__(self, other: object) -> Callable[[Run], bool]:  # type: ignore[override]
        return self._cmp(eq, other)

    def __ne__(self, other: object) -> Callable[[Run], bool]:  # type: ignore[override]
        return self._cmp(ne, other)

    def __lt__(self, other: Any) -> Callable[[Run], bool]:
        return self._cmp(lt, other)

    def __le__(self, other: Any) -> Callable[[Run], bool]:
        return self._cmp(le, other)

    def __gt__(self, other: Any) -> Callable[[Run], bool]:
        return self._cmp(gt, other)

    def __ge__(self, other: Any) -> Callable[[Run], bool]:
        return self._cmp(ge, other)


class AbstractParam(AbstractSelector, ABC): ...


class Id(AbstractParam):
    def __init__(self, label: str | None = None) -> None:
        """Selector for the ID of the run.

        Args:
            label: Optional label for displaying purposes. If None, defaults to `name`.
                Defaults to None.
        """
        super().__init__("id", label)

    def __call__(self, run: Run) -> str:
        return run.id


class Param(AbstractParam):
    """Selector for a specific parameter of the run."""

    def __call__(self, run: Run) -> int | float | str | None:
        return run.params.get(self.name)


class AbstractMetric(AbstractSelector, ABC):
    def __init__(
        self,
        name: str,
        direction: Literal["min", "max"],
        label: str | None = None,
    ) -> None:
        super().__init__(name, label)
        if direction not in ("min", "max"):
            raise ValueError(
                f"Invalid direction: '{direction}'. Must be 'min' or 'max'."
            )
        self.direction = direction


class Metric(AbstractMetric):
    """Selector for a specific metric of the run.

    Args:
        name: Name of the metric to select on.
        direction: Direction of the metric. "min" for minimization, "max" for
            maximization.
        label: Optional label for displaying purposes. If None, defaults to `name`.
            Defaults to None.
    """

    def __call__(self, run: Run) -> float:
        val = run.metrics.get(self.name)
        if val is None:
            return float("-inf") if self.direction == "max" else float("inf")
        return val


class TemporalMetric(AbstractMetric):
    def __init__(
        self,
        name: str,
        direction: Literal["min", "max"],
        reduction: Literal["min", "max", "first", "last"] | None = None,
        label: str | None = None,
    ) -> None:
        """Selector for a specific temporal metric of the run.

        Args:
            name: Name of the temporal metric to select on.
            direction: Direction of the metric. "min" for minimization, "max" for
                maximization.
            reduction: Reduction method to apply to the temporal metric. "min" for
                minimum, "max" for maximum, "first" for the first value, and "last"
                for the last value. If None, the direction is used as the reduction.
                Defaults to None.
            label: Optional label for displaying purposes. If None, defaults to `name`.
                Defaults to None.
        """
        super().__init__(name, direction, label)
        if reduction is not None and reduction not in ("min", "max", "first", "last"):
            raise ValueError(
                f"Invalid reduction method: '{reduction}'. Must be 'min', 'max', "
                "'first', or 'last'."
            )
        self.reduction = reduction or direction

    def __call__(self, run: Run) -> float:
        values = run.temporal.get(self.name, [])
        if not values:
            return float("nan")

        match self.reduction:
            case "min":
                return min(v for _, v in values)
            case "max":
                return max(v for _, v in values)
            case "first":
                return values[0][1]
            case "last":
                return values[-1][1]



from typing import List

import pytest

from ablate.core.types import GroupedRun, Run
from ablate.queries.grouped_query import GroupedQuery
from ablate.queries.query import Query
from ablate.queries.selectors import Metric, Param


@pytest.fixture
def runs() -> List[Run]:
    return [
        Run(id="a", params={"model": "resnet", "seed": 1}, metrics={"accuracy": 0.7}),
        Run(id="b", params={"model": "resnet", "seed": 2}, metrics={"accuracy": 0.8}),
        Run(id="c", params={"model": "vit", "seed": 1}, metrics={"accuracy": 0.6}),
        Run(id="d", params={"model": "vit", "seed": 2}, metrics={"accuracy": 0.9}),
    ]


@pytest.fixture
def grouped(runs: List[Run]) -> GroupedQuery:
    return Query(runs).groupby(Param("model"))


def test_filter_keeps_matching_groups(grouped: GroupedQuery) -> None:
    filtered = grouped.filter(lambda g: g.value == "resnet")
    assert len(filtered) == 1
    assert all(run.params["model"] == "resnet" for run in filtered.all())


def test_map_modifies_each_group(grouped: GroupedQuery) -> None:
    def fn(group: GroupedRun) -> GroupedRun:
        group.runs = [Run(**{**r.model_dump(), "id": r.id + "_x"}) for r in group.runs]
        return group

    mapped = grouped.map(fn)
    assert all(r.id.endswith("_x") for r in mapped.all())


def test_sort_sorts_within_each_group(grouped: GroupedQuery) -> None:
    grouped = grouped.sort(Metric("accuracy", direction="max"), ascending=True)
    for group in grouped._grouped:
        accs = [r.metrics["accuracy"] for r in group.runs]
        assert accs == sorted(accs)


def test_head_tail_topk_bottomk_all_return_expected_shape(
    grouped: GroupedQuery,
) -> None:
    assert len(grouped.head(1).all()) == 2
    assert len(grouped.tail(1).all()) == 2
    assert len(grouped.topk(Metric("accuracy", direction="max"), 1).all()) == 2
    assert len(grouped.bottomk(Metric("accuracy", direction="max"), 1).all()) == 2


def test_aggregate_all_strategies(grouped: GroupedQuery) -> None:
    m = Metric("accuracy", direction="max")
    assert len(grouped.aggregate("first", over=m).all()) == 2
    assert len(grouped.aggregate("last", over=m).all()) == 2
    assert len(grouped.aggregate("best", over=m).all()) == 2
    assert len(grouped.aggregate("worst", over=m).all()) == 2
    assert len(grouped.aggregate("mean", over=m).all()) == 2

    with pytest.raises(ValueError, match="Unsupported aggregation method"):
        grouped.aggregate("unsupported", over=m)  # type: ignore[arg-type]


def test_aggregate_best_worst_missing_over(grouped: GroupedQuery) -> None:
    with pytest.raises(ValueError, match="Method 'best' requires a metric"):
        grouped.aggregate("best")
    with pytest.raises(ValueError, match="Method 'worst' requires a metric"):
        grouped.aggregate("worst")


def test_aggregate_mean_collapses_metadata_and_temporal() -> None:
    run1 = Run(
        id="a",
        params={"model": "resnet", "seed": 1},
        metrics={"acc": 0.8},
        temporal={"acc": [(1, 0.2), (2, 0.6)]},
    )
    run2 = Run(
        id="b",
        params={"model": "resnet", "seed": 2},
        metrics={"acc": 0.4},
        temporal={"acc": [(1, 0.6), (2, 1.0)]},
    )
    grouped = GroupedQuery([GroupedRun(key="model", value="resnet", runs=[run1, run2])])
    agg = grouped.aggregate("mean", over=Metric("acc", direction="max")).all()[0]

    assert agg.params["model"] == "resnet"
    assert agg.params["seed"] == "#"
    assert agg.metrics["acc"] == pytest.approx(0.6)
    assert agg.temporal["acc"] == [(1, 0.4), (2, 0.8)]


def test_to_query_and_all_return_same_runs(grouped: GroupedQuery) -> None:
    assert grouped._to_query().all() == grouped.all()


def test_copy_and_deepcopy(grouped: GroupedQuery) -> None:
    shallow = grouped.copy()
    deep = grouped.deepcopy()

    assert shallow._grouped == grouped._grouped
    assert shallow._grouped is not grouped._grouped

    assert deep._grouped == grouped._grouped
    assert deep._grouped is not grouped._grouped
    assert all(
        dr is not gr for dr, gr in zip(deep._grouped, grouped._grouped, strict=False)
    )


def test_grouped_query_project_reduces_param_space(grouped: GroupedQuery) -> None:
    grouped = grouped.project(Param("model"))
    for group in grouped._grouped:
        for run in group.runs:
            assert set(run.params.keys()) == {"model"}
            assert set(run.metrics.keys()) == {"accuracy"}



from typing import List

from ablate.blocks import AbstractFigureBlock, AbstractTextBlock
from ablate.core.types import Run


class DummyTextBlock(AbstractTextBlock):
    def build(self, runs: List[Run]) -> str:  # type: ignore[override]
        return "not supported"


class DummyFigureBlock(AbstractFigureBlock):
    def build(self, runs: List[Run]) -> str:  # type: ignore[override]
        return "not a dataframe"



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, List

import pandas as pd

from ablate.queries import AbstractMetric, Id, Param

from .abstract_block import AbstractBlock


if TYPE_CHECKING:  # pragma: no cover
    from ablate.core.types import Run


class AbstractFigureBlock(AbstractBlock, ABC):
    @abstractmethod
    def build(self, runs: List[Run]) -> pd.DataFrame: ...


class MetricPlot(AbstractFigureBlock):
    def __init__(
        self,
        metrics: AbstractMetric | List[AbstractMetric],
        identifier: Param | None = None,
        runs: List[Run] | None = None,
    ) -> None:
        """Block for plotting metrics over time.

        Args:
            metrics: Metric or list of metrics to be plotted over time.
            identifier: Optional identifier for the runs. If None, the run ID is used.
                Defaults to None.
            runs: Optional list of runs to be used for the block instead of the default
                runs from the report. Defaults to None.
        """
        super().__init__(runs)
        self.metrics = metrics if isinstance(metrics, list) else [metrics]
        self.identifier = identifier or Id()

    def build(self, runs: List[Run]) -> pd.DataFrame:
        data = []
        for run in runs:
            for metric in self.metrics:
                series = run.temporal.get(metric.name, [])
                for step, value in series:
                    data.append(
                        {
                            "step": step,
                            "value": value,
                            "metric": metric.label,
                            "run": self.identifier(run),
                            "run_id": run.id,
                        }
                    )
        return pd.DataFrame(data)



import datetime
import os
import sys
from typing import Any, Dict, List

from sphinx.application import Sphinx
import tomli


def build_authors(authors: List[Dict[str, str]]) -> str:
    return ", ".join([f"{author['name']}" for author in authors if "name" in author])


with open("../../pyproject.toml", "rb") as f:
    pyproject = tomli.load(f)

sys.path.insert(0, os.path.abspath(".."))
sys.path.append(os.path.abspath("../extensions"))

project = pyproject["project"]["name"]
description = pyproject["project"]["description"]
author = build_authors(pyproject["project"]["authors"])
copyright = f"{datetime.datetime.now().year} (MIT) {author}"
release = pyproject["project"]["version"]


extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx.ext.autosummary",
    "sphinx_autodoc_typehints",
    "sphinx_copybutton",
    "sphinx_design",
    "sphinxarg.ext",
    "sphinxcontrib.jquery",
    "sphinx_lexers",
]

master_doc = "index"
html_theme = "pydata_sphinx_theme"
html_static_path = ["_static"]
html_css_files = {"theme.css"}
html_js_files = ["external_links.js"]
autosummary_generate = True
autodoc_member_order = "bysource"
autoclass_content = "init"

html_theme_options = {
    "show_prev_next": False,
    "footer_start": ["copyright"],
    "footer_end": [],
    "collapse_navigation": False,
    "secondary_sidebar_items": ["page-toc"],
    "navbar_center": [],
    "logo": {
        "alt_text": f"{project} — {description}",
        "text": project,
        "image_dark": "_static/logo_dark.webp",
        "image_light": "_static/logo_light.webp",
    },
    "pygments_light_style": "catppuccin-latte",
    "pygments_dark_style": "catppuccin-mocha",
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/ramppdev/ablate",
            "icon": "fa-brands fa-github",
            "type": "fontawesome",
        },
    ],
}
html_sidebars = {"**": ["globaltoc"]}
html_favicon = "_static/favicon.ico"
html_title = ""


def set_custom_title(
    app: Sphinx,
    pagename: str,
    templatename: str,
    context: Dict[str, Any],
    doctree: Any,
) -> None:
    if pagename == app.config.master_doc:
        context["title"] = f"{project} — {description}"
    elif context.get("title"):
        context["title"] = f"{context['title']} | {project} — {description}"


def setup(app: Sphinx) -> None:
    app.connect("html-page-context", set_custom_title)



from .abstract_block import AbstractBlock
from .figure_blocks import AbstractFigureBlock, MetricPlot
from .table_blocks import AbstractTableBlock, Table
from .text_blocks import H1, H2, H3, H4, H5, H6, AbstractTextBlock, Text


__all__ = [
    "AbstractBlock",
    "AbstractFigureBlock",
    "AbstractTableBlock",
    "AbstractTextBlock",
    "H1",
    "H2",
    "H3",
    "H4",
    "H5",
    "H6",
    "MetricPlot",
    "Table",
    "Text",
]



from typing import List

import pytest

from ablate.core.types import Run
from ablate.queries.query import Query
from ablate.queries.selectors import Metric, Param


@pytest.fixture
def runs() -> List[Run]:
    return [
        Run(id="a", params={"model": "resnet", "seed": 1}, metrics={"accuracy": 0.7}),
        Run(id="b", params={"model": "resnet", "seed": 2}, metrics={"accuracy": 0.8}),
        Run(id="c", params={"model": "vit", "seed": 1}, metrics={"accuracy": 0.9}),
    ]


def test_filter(runs: List[Run]) -> None:
    runs = Query(runs).filter(Param("model") == "resnet").all()
    assert len(runs) == 2
    assert all(r.params["model"] == "resnet" for r in runs)


def test_sort(runs: List[Run]) -> None:
    runs = Query(runs).sort(Metric("accuracy", direction="min")).all()
    assert [r.id for r in runs] == ["c", "b", "a"]


def test_head_tail(runs: List[Run]) -> None:
    q = Query(runs)
    assert q.head(1).all()[0].id == "a"
    assert q.tail(1).all()[0].id == "c"


def test_topk_bottomk(runs: List[Run]) -> None:
    q = Query(runs)
    top = q.topk(Metric("accuracy", direction="max"), k=2).all()
    bot = q.bottomk(Metric("accuracy", direction="max"), k=1).all()
    assert [r.id for r in top] == ["c", "b"]
    assert [r.id for r in bot] == ["a"]


def test_map(runs: List[Run]) -> None:
    q = Query(runs)

    def upper_case_id(run: Run) -> Run:
        run.id = run.id.upper()
        return run

    updated = q.map(upper_case_id).all()
    assert updated[0].id == "A"
    assert runs[0].id == "a"


def test_groupby_single_key(runs: List[Run]) -> None:
    gq = Query(runs).groupby(Param("model"))
    assert len(gq) == 2
    assert {g.value for g in gq._grouped} == {"resnet", "vit"}


def test_groupby_multiple_keys(runs: List[Run]) -> None:
    gq = Query(runs).groupby([Param("model"), Param("seed")])
    keys = {(g.key, g.value) for g in gq._grouped}
    assert len(keys) == 3


def test_groupdiff(runs: List[Run]) -> None:
    grouped = Query(runs).groupdiff(Param("seed"))._grouped
    expected_group_sizes = {"resnet": 2, "vit": 1}
    for group in grouped:
        model = group.runs[0].params["model"]
        assert len(group.runs) == expected_group_sizes[model]
    assert all(len(g.value) == 8 for g in grouped)


def test_query_copy_shallow(runs: List[Run]) -> None:
    original = Query(runs)
    copied = original.copy()

    assert copied._runs == original._runs
    assert copied._runs is not original._runs
    assert copied._runs[0] is original._runs[0]


def test_query_deepcopy(runs: List[Run]) -> None:
    original = Query(runs)
    deepcopied = original.deepcopy()
    assert deepcopied._runs == original._runs
    assert deepcopied._runs is not original._runs
    assert deepcopied._runs[0] is not original._runs[0]


def test_query_len(runs: List[Run]) -> None:
    assert len(Query(runs)) == 3


def test_project_reduces_parameter_space(runs: List[Run]) -> None:
    q = Query(runs).project(Param("model"))
    for run in q.all():
        assert set(run.params.keys()) == {"model"}
        assert set(run.metrics.keys()) == {"accuracy"}



from typing import List

import pandas as pd

from ablate.blocks import H1, MetricPlot, Table, Text
from ablate.core.types import Run
from ablate.queries.selectors import Metric, Param


def make_runs() -> List[Run]:
    return [
        Run(
            id="a",
            params={"model": "resnet", "seed": 1},
            metrics={"accuracy": 0.7},
            temporal={"accuracy": [(0, 0.6), (1, 0.7)]},
        ),
        Run(
            id="b",
            params={"model": "resnet", "seed": 2},
            metrics={"accuracy": 0.8},
            temporal={"accuracy": [(0, 0.7), (1, 0.8)]},
        ),
    ]


def test_text_blocks() -> None:
    assert Text(" simple ").build(make_runs()) == "simple"
    assert H1("# Title").build(make_runs()) == "# Title"


def test_table_block() -> None:
    table = Table(columns=[Param("model"), Param("seed")])
    df = table.build(make_runs())
    assert isinstance(df, pd.DataFrame)
    assert list(df.columns) == ["model", "seed"]
    assert df.iloc[0]["model"] == "resnet"


def test_metric_plot_single() -> None:
    plot = MetricPlot(Metric("accuracy", direction="max"), identifier=Param("seed"))
    df = plot.build(make_runs())
    assert isinstance(df, pd.DataFrame)
    assert set(df.columns) >= {"step", "value", "metric", "run", "run_id"}
    assert df["metric"].unique().tolist() == ["accuracy"]


def test_metric_plot_multi() -> None:
    plot = MetricPlot([Metric("accuracy", direction="max")], identifier=Param("seed"))
    df = plot.build(make_runs())
    assert isinstance(df, pd.DataFrame)
    assert all(k in df.columns for k in ["step", "value", "metric", "run"])



from typing import Any, Dict, List, Tuple

from pydantic import BaseModel


class Run(BaseModel):
    id: str
    params: Dict[str, Any]
    metrics: Dict[str, float]
    temporal: Dict[str, list[Tuple[int, float]]] = {}

    def __init__(
        self,
        id: str,
        params: Dict[str, Any],
        metrics: Dict[str, float],
        temporal: Dict[str, list[Tuple[int, float]]] | None = None,
    ) -> None:  # sphinx needs an explicit __init__ for autodoc
        """A single run of an experiment.

        Args:
            id: Unique identifier for the run.
            params: Parameters used for the run.
            metrics: Metrics recorded during the run.
            temporal: Temporal data recorded during the run. If None, an empty
                dictionary is used. Defaults to None.
        """
        super().__init__(id=id, params=params, metrics=metrics, temporal=temporal or {})


class GroupedRun(BaseModel):
    key: str
    value: str
    runs: List[Run]

    def __init__(
        self,
        key: str,
        value: str,
        runs: List[Run],
    ) -> None:  # sphinx needs an explicit __init__ for autodoc
        """A collection of runs grouped by a key-value pair.

        Args:
            key: Key used to group the runs.
            value: Value used to group the runs.
            runs: List of runs that belong to this group.
        """
        super().__init__(key=key, value=value, runs=runs)



from types import SimpleNamespace
from unittest.mock import MagicMock, patch

import pytest

from ablate.sources import MLflow


@pytest.mark.filterwarnings("ignore::pydantic.PydanticDeprecatedSince20")
@pytest.mark.parametrize(
    ("tracking_uri", "expected_uri_startswith"),
    [
        (None, None),
        ("/some/local/path", "file://"),
        ("http://mlflow.mycompany.com", "http://"),
        ("https://mlflow.mycompany.com", "https://"),
        ("file:///already/uri", "file://"),
    ],
)
@patch("mlflow.tracking.MlflowClient")
def test_mlflow_uri_resolution(
    client: MagicMock,
    tracking_uri: str,
    expected_uri_startswith: str,
) -> None:
    mock_client = client.return_value
    mock_run = SimpleNamespace(
        info=SimpleNamespace(run_id="run-1"),
        data=SimpleNamespace(
            params={"lr": "0.01"},
            metrics={"accuracy": 0.9},
            tags={"mlflow.runName": "example"},
        ),
    )
    mock_client.get_experiment_by_name.return_value = SimpleNamespace(
        experiment_id="123"
    )
    mock_client.search_runs.return_value = [mock_run]
    mock_client.get_metric_history.return_value = [SimpleNamespace(step=1, value=0.9)]

    source = MLflow(tracking_uri=tracking_uri, experiment_names="default")
    runs = source.load()

    if expected_uri_startswith is None:
        client.assert_called_once_with()
    else:
        uri_arg = client.call_args.args[0]
        assert uri_arg.startswith(expected_uri_startswith)

    r = runs[0]
    assert r.id == "run-1"
    assert r.params == {"lr": "0.01", "mlflow.runName": "example"}
    assert r.metrics == {"accuracy": 0.9}
    assert r.temporal == {"accuracy": [(1, 0.9)]}


@patch.dict("sys.modules", {"mlflow.tracking": None})
def test_import_error_if_mlflow_not_installed() -> None:
    with pytest.raises(ImportError, match="MLflow source requires `mlflow`"):
        MLflow(tracking_uri="/fake", experiment_names=["exp"])


@patch("mlflow.tracking.MlflowClient")
def test_mlflow_raises_on_invalid_experiment_name(client: MagicMock) -> None:
    mock_client = client.return_value
    mock_client.get_experiment_by_name.return_value = None

    source = MLflow(tracking_uri="/fake/path", experiment_names=["nonexistent"])

    with pytest.raises(ValueError, match="One or more experiment names not found"):
        source.load()



from pathlib import Path
from typing import List
from urllib.parse import urlparse

from ablate.core.types import Run

from .abstract_source import AbstractSource


class MLflow(AbstractSource):
    def __init__(
        self,
        experiment_names: str | List[str],
        tracking_uri: str | None,
    ) -> None:
        """MLflow source for loading runs from a MLflow server.

        Args:
            experiment_names: An experiment name or a list of experiment names to load
                runs from.
            tracking_uri: The URI or local path to the MLflow tracking server.
                If None, use the default tracking URI set in the MLflow configuration.
                Defaults to None.

        Raises:
            ImportError: If the `mlflow` package is not installed.
        """
        try:
            from mlflow.tracking import MlflowClient
        except ImportError as e:
            raise ImportError(
                "MLflow source requires `mlflow`. "
                "Install via `pip install ablate[mlflow]`."
            ) from e

        self.tracking_uri = tracking_uri
        self.experiment_names = (
            [experiment_names]
            if isinstance(experiment_names, str)
            else experiment_names
        )
        if not tracking_uri:
            self.client = MlflowClient()
            return
        if urlparse(tracking_uri).scheme in {"http", "https", "file"}:
            uri = tracking_uri
        else:
            uri = Path(tracking_uri).resolve().as_uri()
        self.client = MlflowClient(uri)

    def load(self) -> List[Run]:
        ids = [self.client.get_experiment_by_name(n) for n in self.experiment_names]
        if not all(ids):
            raise ValueError(
                f"One or more experiment names not found: {self.experiment_names}"
            )
        runs = self.client.search_runs([e.experiment_id for e in ids if e])
        records: List[Run] = []
        for run in runs:
            p, m, t = run.data.params, run.data.metrics, {}
            p.update(run.data.tags)
            for name in m:
                history = self.client.get_metric_history(run.info.run_id, name)
                t[name] = [(h.step, h.value) for h in history]
            records.append(Run(id=run.info.run_id, params=p, metrics=m, temporal=t))
        return records



from pathlib import Path
from types import SimpleNamespace
from unittest.mock import MagicMock, patch

import pytest

from ablate.sources import TensorBoard


@patch.dict(
    "sys.modules",
    {"tensorboard.backend.event_processing.event_accumulator": None},
)
def test_import_error_if_tensorboard_not_installed() -> None:
    with pytest.raises(ImportError, match="TensorBoard source requires `tensorboard`"):
        TensorBoard(logdirs="/some/path")


@patch("tensorboard.backend.event_processing.event_accumulator.EventAccumulator")
def test_tensorboard_load_single_run(
    mock_event_accumulator: MagicMock, tmp_path: Path
) -> None:
    run_dir = tmp_path / "run1"
    run_dir.mkdir()
    (run_dir / "events.out.tfevents.12345").touch()

    mock_ea_instance = mock_event_accumulator.return_value
    mock_ea_instance.Tags.return_value = {"scalars": ["accuracy", "loss"]}
    mock_ea_instance.Scalars.side_effect = lambda tag: [
        SimpleNamespace(step=1, value=0.5 if tag == "accuracy" else 0.9)
    ]
    mock_ea_instance.Reload.return_value = None

    source = TensorBoard(logdirs=str(tmp_path))
    runs = source.load()

    assert len(runs) == 1
    run = runs[0]
    assert run.id == "run1"
    assert run.params == {}
    assert run.metrics == {"accuracy": 0.5, "loss": 0.9}
    assert run.temporal == {
        "accuracy": [(1, 0.5)],
        "loss": [(1, 0.9)],
    }


@patch("tensorboard.backend.event_processing.event_accumulator.EventAccumulator")
def test_tensorboard_load_multiple_runs(
    mock_event_accumulator: MagicMock, tmp_path: Path
) -> None:
    for name in ["runA", "runB"]:
        path = tmp_path / name
        path.mkdir()
        (path / "events.out.tfevents.12345").touch()

    mock_ea_instance = mock_event_accumulator.return_value
    mock_ea_instance.Tags.return_value = {"scalars": ["acc"]}
    mock_ea_instance.Scalars.return_value = [SimpleNamespace(step=1, value=0.75)]
    mock_ea_instance.Reload.return_value = None

    source = TensorBoard(logdirs=[str(tmp_path)])
    runs = source.load()

    run_ids = sorted(r.id for r in runs)
    assert run_ids == ["runA", "runB"]
    for run in runs:
        assert run.metrics["acc"] == 0.75
        assert run.temporal["acc"] == [(1, 0.75)]


@patch("tensorboard.backend.event_processing.event_accumulator.EventAccumulator")
def test_tensorboard_handles_empty_logs(
    mock_event_accumulator: MagicMock, tmp_path: Path
) -> None:
    logdir = tmp_path / "empty_run"
    logdir.mkdir()
    (logdir / "events.out.tfevents.12345").touch()

    mock_ea_instance = mock_event_accumulator.return_value
    mock_ea_instance.Tags.return_value = {"scalars": []}
    mock_ea_instance.Reload.return_value = None

    source = TensorBoard(logdirs=str(tmp_path))
    runs = source.load()

    assert len(runs) == 1
    assert runs[0].metrics == {}
    assert runs[0].temporal == {}



import pytest

from ablate.core.types import Run
from ablate.queries.selectors import Id, Metric, Param, TemporalMetric


@pytest.fixture
def example_run() -> Run:
    return Run(
        id="run-42",
        params={"model": "resnet", "lr": 0.001},
        metrics={"accuracy": 0.91, "loss": 0.1},
        temporal={"accuracy": [(0, 0.5), (1, 0.8), (2, 0.9)]},
    )


def test_id_selector(example_run: Run) -> None:
    selector = Id()
    assert selector(example_run) == "run-42"
    assert selector(example_run) != "run-x"


def test_param_selector(example_run: Run) -> None:
    selector = Param("lr")
    assert selector(example_run) == 0.001
    assert (selector > 0.0001)(example_run)
    assert not (selector < 0.0001)(example_run)

    missing_selector = Param("missing")
    assert missing_selector(example_run) is None


def test_param_comparisons(example_run: Run) -> None:
    selector = Param("lr")
    assert (selector == 0.001)(example_run)
    assert not (selector == 0.01)(example_run)

    assert (selector != 0.01)(example_run)
    assert not (selector != 0.001)(example_run)

    assert (selector <= 0.001)(example_run)
    assert (selector >= 0.001)(example_run)
    assert not (selector <= 0.0001)(example_run)
    assert not (selector >= 0.01)(example_run)


def test_metric_selector(example_run: Run) -> None:
    selector = Metric("accuracy", direction="max")
    assert selector(example_run) == 0.91
    assert (selector > 0.5)(example_run)
    assert not (selector < 0.5)(example_run)

    missing = Metric("missing", direction="max")
    assert missing(example_run) == float("-inf")


def test_invalid_metric_direction() -> None:
    with pytest.raises(ValueError, match="Invalid direction"):
        Metric("accuracy", direction="invalid")  # type: ignore[arg-type]


@pytest.mark.parametrize(
    ("reduction", "expected"),
    [
        ("min", 0.5),
        ("max", 0.9),
        ("first", 0.5),
        ("last", 0.9),
    ],
)
def test_temporal_metric_selector(
    example_run: Run, reduction: str, expected: float
) -> None:
    selector = TemporalMetric("accuracy", direction="max", reduction=reduction)  # type: ignore[arg-type]
    assert selector(example_run) == expected


def test_temporal_metric_missing_returns_nan(example_run: Run) -> None:
    selector = TemporalMetric("not_logged", direction="max", reduction="max")
    assert selector(example_run) != selector(example_run)


def test_temporal_metric_invalid_reduction() -> None:
    with pytest.raises(ValueError, match="Invalid reduction method"):
        TemporalMetric("accuracy", direction="max", reduction="median")  # type: ignore[arg-type]



import tomli

import ablate


def test_version() -> None:
    with open("pyproject.toml", "rb") as f:
        pyproject = tomli.load(f)
    assert pyproject["project"]["version"] == ablate.__version__



from pathlib import Path
from typing import List

from ablate.core.types import Run

from .abstract_source import AbstractSource


class TensorBoard(AbstractSource):
    def __init__(self, logdirs: str | List[str]) -> None:
        """TensorBoard source for loading runs from event logs.

        Args:
            logdirs: A path or list of paths to TensorBoard event log directories.
        """
        try:
            from tensorboard.backend.event_processing import (
                event_accumulator,  # noqa: F401
            )
        except ImportError as e:
            raise ImportError(
                "TensorBoard source requires `tensorboard`. "
                "Please install with `pip install ablate[tensorboard]`."
            ) from e

        self.logdirs = (
            [Path(logdirs)] if isinstance(logdirs, str) else [Path(p) for p in logdirs]
        )

    def load(self) -> List[Run]:
        from tensorboard.backend.event_processing.event_accumulator import (
            EventAccumulator,
        )

        records: List[Run] = []

        for logdir in self.logdirs:
            for path in logdir.glob("**/events.out.tfevents.*"):
                ea = EventAccumulator(str(path.parent))
                ea.Reload()

                metrics = {}
                temporal = {}

                for tag in ea.Tags().get("scalars", []):
                    scalar_events = ea.Scalars(tag)
                    if scalar_events:
                        last_value = scalar_events[-1].value
                        metrics[tag] = last_value
                        temporal[tag] = [(e.step, e.value) for e in scalar_events]

                run_id = path.parent.name  # use folder name as ID

                records.append(
                    Run(id=run_id, params={}, metrics=metrics, temporal=temporal)
                )

        return records



from abc import ABC, abstractmethod
from typing import List

import pandas as pd

from ablate.core.types import Run
from ablate.queries import AbstractSelector

from .abstract_block import AbstractBlock


class AbstractTableBlock(AbstractBlock, ABC):
    def __init__(
        self,
        columns: List[AbstractSelector],
        runs: List[Run] | None = None,
    ) -> None:
        """Table block for a report.

        Args:
            columns: Columns to be included in the table. Each column is defined by a
                selector that extracts the data from the runs.
            runs: Optional list of runs to be used for the block instead of the default
                runs from the report. Defaults to None.
        """
        super().__init__(runs)
        self.columns = columns

    @abstractmethod
    def build(self, runs: List[Run]) -> pd.DataFrame: ...


class Table(AbstractTableBlock):
    def build(self, runs: List[Run]) -> pd.DataFrame:
        rows = []
        for run in runs:
            row = {}
            for column in self.columns:
                row[column.label] = column(run)
            rows.append(row)
        return pd.DataFrame(rows, columns=[column.label for column in self.columns])



import sys
from typing import List

from matplotlib import pyplot as plt

from ablate.blocks import (
    AbstractFigureBlock,
    AbstractTableBlock,
    AbstractTextBlock,
    MetricPlot,
    Text,
)
from ablate.blocks.text_blocks import _Heading
from ablate.core.types import Run
from ablate.exporters.abstract_exporter import AbstractExporter
from ablate.exporters.utils import HEADING_LEVELS
from ablate.report import Report

from .utils import create_metric_plot


def running_in_notebook() -> bool:
    return any("jupyter" in arg or "ipykernel" in arg for arg in sys.argv)


class Notebook(AbstractExporter):
    def __init__(self) -> None:
        super().__init__()
        try:
            from IPython.display import display  # noqa: F401
        except ImportError as e:
            raise ImportError(
                "Notebook exporter requires `jupyter`. "
                "Please install with `pip install ablate[jupyter]`."
            ) from e

    def export(self, report: Report) -> None:
        if not running_in_notebook():
            raise RuntimeError(
                "Notebook exporter can only be used inside a Jupyter notebook."
            )
        self.render_blocks(report)

    def render_text(self, block: AbstractTextBlock, runs: List[Run]) -> None:
        from IPython.display import Markdown, display

        if isinstance(block, Text):
            display(Markdown(block.build(runs)))
        elif isinstance(block, _Heading):
            level = HEADING_LEVELS[type(block)]
            display(Markdown(f"{'#' * level} {block.build(runs)}"))
        else:
            raise NotImplementedError(f"Unsupported text block: '{type(block)}'")

    def render_table(self, block: AbstractTableBlock, runs: List[Run]) -> None:
        from IPython.display import display

        display(block.build(runs))

    def render_figure(self, block: AbstractFigureBlock, runs: List[Run]) -> None:
        from IPython.display import Markdown, display

        if not isinstance(block, MetricPlot):
            raise NotImplementedError(f"Unsupported figure block: '{type(block)}'.")

        df = block.build(runs)
        if df.empty:
            m = f"*No data available for {', '.join(m.label for m in block.metrics)}*"
            display(Markdown(m))
            return

        fig = create_metric_plot(df, block.identifier.label)
        plt.close(fig)
        display(fig)



from __future__ import annotations

from collections import defaultdict
from copy import deepcopy
from typing import TYPE_CHECKING, Callable, Dict, List, Literal, Union

from ablate.core.types import GroupedRun, Run


if TYPE_CHECKING:  # pragma: no cover
    from .query import Query  # noqa: TC004
    from .selectors import AbstractMetric, AbstractParam


class GroupedQuery:
    def __init__(self, groups: List[GroupedRun]) -> None:
        """Query interface for manipulating grouped runs in a functional way.

        All methods operate on a shallow copy of the runs in the query, so the original
        runs are not modified and assumed to be immutable.

        Args:
            groups: A list of grouped runs to be queried.
        """
        self._grouped = groups

    def filter(self, fn: Callable[[GroupedRun], bool]) -> GroupedQuery:
        """Filter the grouped runs in the grouped query based on a predicate function.

        Args:
            fn: Predicate function that takes in a grouped run and returns a boolean
                value.

        Returns:
            A new grouped query with the grouped runs that satisfy the predicate
            function.
        """
        return GroupedQuery([g for g in self._grouped[:] if fn(g)])

    def map(self, fn: Callable[[GroupedRun], GroupedRun]) -> GroupedQuery:
        """Apply a function to each grouped run in the grouped query.

        This function is intended to be used for modifying the grouped runs in the
        grouped query. The function should return a new grouped run object as the
        original grouped run is not modified.

        Args:
            fn: Function that takes in a grouped run and returns a new grouped run
                object.

        Returns:
            A new grouped query with the modified grouped runs.
        """
        return GroupedQuery([fn(deepcopy(g)) for g in self._grouped])

    def sort(self, key: AbstractMetric, ascending: bool = False) -> GroupedQuery:
        """Sort the runs inside each grouped run in the grouped query based on a metric.

        Args:
            key: Metric to sort the grouped runs by.
            ascending: Whether to sort in ascending order.
                Defaults to False (descending order).

        Returns:
            A new grouped query with the grouped runs sorted by the specified metric.
        """
        return GroupedQuery(
            [
                GroupedRun(
                    key=g.key,
                    value=g.value,
                    runs=sorted(g.runs, key=key, reverse=not ascending),
                )
                for g in self._grouped
            ]
        )

    def project(
        self, selectors: Union[AbstractParam, List[AbstractParam]]
    ) -> GroupedQuery:
        """Project the parameter space of the grouped runs in the grouped query to a
        subset of parameters only including the specified selectors.

        This function is intended to be used for reducing the dimensionality of the
        parameter space and therefore operates on a deep copy of the grouped runs in the
        grouped query.

        Args:
            selectors: Selector or list of selectors to project the grouped runs by.

        Returns:
            A new grouped query with the projected grouped runs.
        """
        if not isinstance(selectors, list):
            selectors = [selectors]

        names = {s.name for s in selectors}
        projected: List[GroupedRun] = []

        for group in deepcopy(self._grouped):
            for run in group.runs:
                run.params = {k: v for k, v in run.params.items() if k in names}
            projected.append(group)

        return GroupedQuery(projected)

    def head(self, n: int) -> Query:
        """Get the first n runs inside each grouped run.

        Args:
            n: Number of runs to return per group.

        Returns:
            A new query with the first n runs from each grouped run.
        """
        return GroupedQuery(
            [
                GroupedRun(key=g.key, value=g.value, runs=g.runs[:n])
                for g in self._grouped
            ]
        )._to_query()

    def tail(self, n: int) -> Query:
        """Get the last n runs inside each grouped run.

        Args:
            n: Number of runs to return per group.

        Returns:
            A new query with the last n runs from each grouped run.
        """
        return GroupedQuery(
            [
                GroupedRun(key=g.key, value=g.value, runs=g.runs[-n:])
                for g in self._grouped
            ]
        )._to_query()

    def topk(self, metric: AbstractMetric, k: int) -> Query:
        """Get the top k runs inside each grouped run based on a metric.

        Args:
            metric: Metric to sort the runs by.
            k: Number of top runs to return per group.

        Returns:
            A new query with the top k runs from each grouped run based on the
            specified metric.
        """
        return GroupedQuery(
            [
                GroupedRun(
                    key=g.key,
                    value=g.value,
                    runs=sorted(g.runs, key=metric, reverse=metric.direction == "min")[
                        :k
                    ],
                )
                for g in self._grouped
            ]
        )._to_query()

    def bottomk(self, metric: AbstractMetric, k: int) -> Query:
        """Get the bottom k runs inside each grouped run based on a metric.

        Args:
            metric: Metric to sort the runs by.
            k: Number of bottom runs to return per group.

        Returns:
            A new query with the bottom k runs from each grouped run based on the
            specified metric.
        """
        return GroupedQuery(
            [
                GroupedRun(
                    key=g.key,
                    value=g.value,
                    runs=sorted(g.runs, key=metric, reverse=metric.direction == "max")[
                        :k
                    ],
                )
                for g in self._grouped
            ]
        )._to_query()

    def aggregate(
        self,
        method: Literal["first", "last", "best", "worst", "mean"],
        over: AbstractMetric | None = None,
    ) -> Query:
        """Aggregate each group of runs using a specified method.

        Supported methods include:
          * :attr:`"first"`: Selects the first run from each group.
          * :attr:`"last"`: Selects the last run from each group.
          * :attr:`"best"`: Selects the run with the best value based on the given
            metric.
          * :attr:`"worst"`: Selects the run with the worst value based on the given
            metric.
          * :attr:`"mean"`: Computes the mean run across all runs in each group,
            including averaged metrics and temporal data, and collapsed metadata.

        Args:
            method: Aggregation strategy to apply per group.
            over: The metric used for comparison when using "best" or "worst" methods.
                Has no effect for "first", "last", or "mean" methods.
                Defaults to None.

        Raises:
            ValueError: If an unsupported aggregation method is provided or if the
                "best" or "worst" method is used without a specified metric.


        Returns:
            A new query with the aggregated runs from each group.
        """
        if method in {"best", "worst"} and over is None:
            raise ValueError(
                f"Method '{method}' requires a metric to be specified for comparison."
            )
        from .query import Query

        match method:
            case "first":
                return self.head(1)
            case "last":
                return self.tail(1)
            case "best":
                assert over is not None
                return self.topk(over, 1)
            case "worst":
                assert over is not None
                return self.bottomk(over, 1)
            case "mean":
                return Query([self._mean_run(g) for g in self._grouped])
            case _:
                raise ValueError(
                    f"Unsupported aggregation method: '{method}'. Must be "
                    "'first', 'last', 'best', 'worst', or 'mean'."
                )

    @staticmethod
    def _mean_run(group: GroupedRun) -> Run:
        def _mean(values: List[float]) -> float:
            return sum(values) / len(values) if values else float("nan")

        def _mean_temporal(runs: List[Run]) -> Dict[str, List[tuple[int, float]]]:
            all_keys = set().union(*(r.temporal.keys() for r in runs))
            step_accumulator: Dict[str, Dict[int, List[float]]] = {}

            for key in all_keys:
                step_values = defaultdict(list)
                for run in runs:
                    for step, val in run.temporal.get(key, []):
                        step_values[step].append(val)
                step_accumulator[key] = step_values

            return {
                key: sorted(
                    (step, sum(vals) / len(vals)) for step, vals in step_values.items()
                )
                for key, step_values in step_accumulator.items()
            }

        def _common_metadata(attr: str) -> Dict[str, str]:
            key_sets = [getattr(r, attr).keys() for r in group.runs]
            common_keys = set.intersection(*map(set, key_sets))
            result = {}
            for k in common_keys:
                values = {str(getattr(r, attr)[k]) for r in group.runs}
                result[k] = next(iter(values)) if len(values) == 1 else "#"
            return result

        all_metrics = [r.metrics for r in group.runs]
        all_keys = set().union(*all_metrics)
        mean_metrics = {
            k: _mean([m[k] for m in all_metrics if k in m]) for k in all_keys
        }

        return Run(
            id=f"grouped:{group.key}:{group.value}",
            params=_common_metadata("params"),
            metrics=mean_metrics,
            temporal=_mean_temporal(group.runs),
        )

    def _to_query(self) -> Query:
        from .query import Query

        return Query([run for group in self._grouped for run in group.runs])

    def all(self) -> List[Run]:
        """Collect all runs in the grouped query by flattening the grouped runs.

        Returns:
            A list of all runs in the grouped query.
        """
        return deepcopy(self._to_query()._runs)

    def copy(self) -> GroupedQuery:
        """Obtain a shallow copy of the grouped query.

        Returns:
            A new grouped query with the same grouped runs as the original grouped
            query.
        """
        return GroupedQuery(self._grouped[:])

    def deepcopy(self) -> GroupedQuery:
        """Obtain a deep copy of the grouped query.

        Returns:
            A new grouped query with deep copies of the grouped runs in the original
            grouped query.
        """
        return GroupedQuery(deepcopy(self._grouped))

    def __len__(self) -> int:
        """Get the number of grouped runs in the grouped query.

        Returns:
            The number of grouped runs in the grouped query.
        """
        return len(self._grouped)



from abc import ABC, abstractmethod
from typing import Any, Callable, List, cast

from ablate.blocks import (
    AbstractBlock,
    AbstractFigureBlock,
    AbstractTableBlock,
    AbstractTextBlock,
)
from ablate.core.types import Run
from ablate.report import Report


class AbstractExporter(ABC):
    @abstractmethod
    def export(self, report: Report) -> None:
        """Export the report.

        Should call :meth:`~ablate.exporters.AbstractExporter.render_blocks` to generate
        the content of the report.

        Args:
            report: The report to be exported.
        """

    def render_blocks(self, report: Report) -> List[Any]:
        """Render a blocks of the report.

        Args:
            report: The report to be rendered.

        Raises:
            ValueError: If the block type is not supported.

        Returns:
            List of rendered blocks.
        """
        render_map = {
            AbstractTextBlock: self.render_text,
            AbstractTableBlock: self.render_table,
            AbstractFigureBlock: self.render_figure,
        }
        content = []
        for block in report.blocks:
            for block_type, render_fn in render_map.items():
                if isinstance(block, block_type):
                    fn = cast("Callable[[AbstractBlock, List[Run]], Any]", render_fn)
                    content.append(self._apply_render_fn(block, fn, report.runs))
                    break
            else:
                raise ValueError(f"Unknown block type: '{type(block)}'.")
        return content

    @staticmethod
    def _apply_render_fn(
        block: AbstractBlock,
        fn: Callable[[AbstractBlock, List[Run]], Any],
        runs: List[Run],
    ) -> Any:
        if block.runs:
            return fn(block, block.runs)
        return fn(block, runs)

    @abstractmethod
    def render_text(self, block: AbstractTextBlock, runs: List[Run]) -> Any:
        """Render a text block.

        Args:
            block: The text block to be rendered.
            runs: The list of runs to be used for the block.

        Returns:
            The rendered text block.
        """

    @abstractmethod
    def render_table(self, block: AbstractTableBlock, runs: List[Run]) -> Any:
        """Render a table block.

        Args:
            block: The table block to be rendered.
            runs: The list of runs to be used for the block.

        Returns:
            The rendered table block.
        """

    @abstractmethod
    def render_figure(self, block: AbstractFigureBlock, runs: List[Run]) -> Any:
        """Render a figure block.

        Args:
            block: The figure block to be rendered.
            runs: The list of runs to be used for the block.

        Returns:
            The rendered figure block.
        """



from abc import ABC, abstractmethod
from typing import Any, List

from ablate.core.types import Run


class AbstractBlock(ABC):
    def __init__(self, runs: List[Run] | None = None) -> None:
        """Abstract content block for a report.

        Args:
            runs: Optional list of runs to be used for the block instead of the default
                runs from the report. Defaults to None.
        """
        self.runs = runs

    @abstractmethod
    def build(self, runs: List[Run]) -> Any:
        """Build the intermediate representation of the block, ready for rendering
        and export.

        Args:
            runs: List of runs to be used for the block.

        Returns:
            The intermediate representation of the block.
        """



from pathlib import Path
from typing import Any, Dict, Generator, List, Union

import pandas as pd
import yaml

from ablate.core.types import Run

from .abstract_source import AbstractSource


def extract_metric_values(metrics: Dict[str, Any], prefix: str = "") -> Dict[str, Any]:
    prefix = f"{prefix}_" if prefix else ""
    return {f"{prefix}{k}": v["all"] for k, v in metrics.items() if k != "iteration"}


def flatten_autrainer_config(
    config: Union[Dict, list],
    prefix: str = "",
) -> Generator[tuple[str, Any], None, None]:
    if isinstance(config, dict):
        id_value = config.get("id")
        if "id" in config and len(config) == 1:
            yield prefix.rstrip("."), id_value
            return
        for key, value in config.items():
            if key == "id":
                yield prefix.rstrip("."), value
                continue
            yield from flatten_autrainer_config(value, f"{prefix}{key}.")
    elif isinstance(config, list):
        for idx, item in enumerate(config):
            yield from flatten_autrainer_config(item, f"{prefix}{idx}.")
    else:
        yield prefix.rstrip("."), config


class Autrainer(AbstractSource):
    def __init__(self, results_dir: str, experiment_id: str) -> None:
        """Autrainer source for loading runs from an autrainer experiment.

        Analogous to `autrainer`, all metrics are reported at the iteration where the
        tracking metric reaches its best development value.

        Args:
            results_dir: The directory where autrainer results are stored.
            experiment_id: The ID of the autrainer experiment to load.

        Raises:
            FileNotFoundError: If the specified experiment ID does not exist in the
                results directory.
        """
        self.results_dir = results_dir
        self.experiment_id = experiment_id
        self._location = Path(results_dir) / experiment_id / "training"
        if not self._location.exists():
            raise FileNotFoundError(
                "No autrainer experiment found with ID "
                f"'{experiment_id}' in directory '{results_dir}'."
            )

    def _load_run(self, path: Path) -> Run:
        run_id = path.name

        with open(path / ".hydra" / "config.yaml") as f:
            params = dict(flatten_autrainer_config(yaml.safe_load(f)))

        with open(path / "_best" / "dev.yaml") as f:
            dev_metrics = extract_metric_values(yaml.safe_load(f))
        with open(path / "_test" / "test_holistic.yaml") as f:
            test_metrics = extract_metric_values(yaml.safe_load(f), "test")
        metrics = {**dev_metrics, **test_metrics}

        df = pd.read_csv(path / "metrics.csv")
        temporal = {
            col: list(zip(df["iteration"], df[col], strict=False))
            for col in df.columns
            if col != "iteration"
        }

        return Run(id=run_id, params=params, metrics=metrics, temporal=temporal)

    def load(self) -> List[Run]:
        runs = []
        for p in Path(self._location).iterdir():
            if not p.is_dir():
                continue
            runs.append(self._load_run(p))
        return runs



from __future__ import annotations

from typing import TYPE_CHECKING, List

from typing_extensions import Self


if TYPE_CHECKING:  # pragma: no cover
    from ablate.blocks import AbstractBlock
    from ablate.core.types import Run


class Report:
    def __init__(self, runs: List[Run]) -> None:
        """Report mapping a list of runs to a list of blocks.

        Args:
            runs: List of runs to be associated with the report.
        """
        self.runs = runs
        self.blocks: List[AbstractBlock] = []

    def add(self, *blocks: AbstractBlock) -> Self:
        """Add one or more blocks to the report.

        Args:
            blocks: One or more blocks to be added to the report.

        Returns:
            The updated report with the added blocks.
        """
        for block in blocks:
            self.blocks.append(block)
        return self

    def __iadd__(self, block: AbstractBlock) -> Self:
        self.blocks.append(block)
        return self

    def __add__(self, block: AbstractBlock) -> Report:
        r = Report(self.runs)
        r.blocks = self.blocks + [block]
        return r
